# zib -- zzznowman's IRC bot
# Copyright (C) 2006  Fabian Pietsch <fabian@zzznowman.dyndns.org>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Channel objects containment

package ZIB::ChannelList;
use strict;
use warnings;
use vars qw($VERSION);
use ZIB qw(:logging);

$VERSION = 20060605.1;

sub new($$)
{
	my ($class, $connection) = @_;

	my $self = {
	  connection => $connection,
	  channels   => {}
	};
	return bless($self, $class);
}

sub conn($) { return $_[0]->{connection} }

sub chans($;@)
{
	my ($self, @channames) = @_;
	my $csref = $self->{channels};

	# FIXME: Really works as expected? Should return hash slice.
	return @$csref{@channames}
	  if (@channames);

	return (wantarray() ? values(%{$csref}) : $csref);
}

sub chan($$)
{
	my ($self, $name) = @_;

	return $self->chans()->{$name};
}

# FIXME
# sub _chans_matching($%)
# {
# 	my ($self, $attrs) = @_;
# 	my @matches;
# 
# CHAN:	foreach my $chan ($self->chans())
# 	{
# 		foreach my $attr (keys(%$attrs))
# 		{
# 			next CHAN unless ($chan->{$attr} eq %$attrs{$attr});
# 		}
# 
# 		push(@matches, $chan);
# 	}
# 
# 	return @matches;
# }
# 
# sub want($)
# {
# 	my ($self) = @_;
# 	return $self->_chans_matching({want => 1});
# }
# 
# sub have($)
# {
# 	my ($self) = @_;
# 	return $self->_chans_matching({have => 1});
# }

sub want($)
{
	my ($self) = @_;
	return grep { $_->want() } ($self->chans());
}

sub have($)
{
	my ($self) = @_;
	return grep { $_->have() } ($self->chans());
}

sub add_chans($@)
{
	my ($self, @newchans) = @_;
	my $csref = $self->chans();
	my @curchans;

	foreach my $chan (@newchans)
	{
		my $isobj = UNIVERSAL::isa($chan, 'ZIB::Channel');
		my $channame = ($isobj ? $chan->name() : $chan);

		# Already in channel list?
		if (defined(my $oldchan = $csref->{$channame}))
		{
			warning('Warning: can\'t add to channel list: another channel object for %s already in list', $channame)
			  if ($isobj && $chan != $oldchan);

			# Just add to returned array.
			push(@curchans, $oldchan);
		}
		else
		{
			# Auto-create a channel object around passed data, if necessary.
			$chan = ZIB::Channel->new($chan)
			  unless ($isobj);

			# Really add to channel list & add to returned array.
			$csref->{$channame} = $chan;
			push(@curchans, $chan);
		}
	}

	return @curchans;
}

sub add_chan($$)
{
	my ($self, $newchan) = @_;
	my @chans = $self->add_chans($newchan);
	return (@chans ? $chans[0] : undef);
}

sub remove_chans($@)
{
	my ($self, @oldchans) = @_;
	my $csref = $self->chans();
	my @removedchans;

	foreach my $chan (@oldchans)
	{
		my $isobj = UNIVERSAL::isa($chan, 'ZIB::Channel');
		my $channame = ($isobj ? $chan->name() : $chan);
		my $candidate = $csref->{$channame};

		unless (defined($candidate))
		{
			warning('Warning: can\'t remove from channel list: channel %s not in list',
			        $channame);
			next;
		}

		if ($isobj && $chan != $candidate)
		{
			warning('Warning: can\'t remove from channel list: channel object for %s in list is another one',
				$channame);
			next;
		}

		push(@removedchans, $candidate);
		delete($csref->{$channame});
	}

	return @removedchans;
}

sub remove_chan($$)
{
	my ($self, $oldchan) = @_;
	my @chans = $self->remove_chans($oldchan);
	return (@chans ? $chans[0] : undef);
}

sub clear($)
{
	my ($self) = @_;
	my $csref = $self->chans();

	delete($csref->{keys %$csref});
}

sub join_chans($@)
{
	my ($self, @newchans) = @_;
	my $conn = $self->conn();
	my @chans = $self->add_chans(@newchans);
	my @joinchannames;

	foreach my $chan (@chans)
	{
		my $channame = $chan->name();

		unless ($chan->want())
		{
			info("Adding channel %s to wantlist", $channame);
			$chan->want(1);
		}

		warning("Warning: trying to join already joined channel %s", $channame)
		  if ($chan->have());

		push(@joinchannames, $channame);
	}

	return @chans unless (@joinchannames);

	info("Trying to join channel%s %s",
	     $#joinchannames == 0 ? '' : 's', join(',', @joinchannames));
	foreach my $channame (@joinchannames)
	{
		$conn->join($channame);
	}

	return @chans;
}

sub join_chan($$)
{
	my ($self, $newchan) = @_;
	my @chans = $self->join_chans($newchan);
	return (@chans ? $chans[0] : undef);
}

sub part_chans($@)
{
	my ($self, @oldchans) = @_;
	my $conn = $self->conn();
	my @chans =
	  map { UNIVERSAL::isa($_, 'ZIB::Channel') ? $_ : $self->chan($_) }
	      (@oldchans);
	my @channames = map { $_->name() } (@chans);

	return @chans unless (@channames);

	info("Trying to part channel%s %s",
	     $#channames == 0 ? '' : 's', join(',', @channames));
	foreach my $chan (@chans)
	{
		my $channame = $chan->name();

		warning('Warning: trying to part unjoined channel %s', $channame)
		  unless ($chan->have());

		$chan->want(0);
		$conn->part($chan->name());
	}

	return @chans;
}

sub part_chan($$)
{
	my ($self, $oldchan) = @_;
	my @chans = $self->part_chans($oldchan);
	return (@chans ? $chans[0] : undef);
}

sub part_all_chans($)
{
	my ($self) = @_;

	info("About to part all channels");
	$self->part_chans(values(%{$self->chans()}));
}

1;
