#!/usr/bin/perl

# zib -- zzznowman's IRC bot
# Copyright (C) 2006  Fabian Pietsch <fabian@zzznowman.dyndns.org>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Version 0.2.3 (20060605-1)

use strict;
use warnings;
use Carp;
use Net::IRC;
use Socket;
use ZIB qw(:all_logging :conversions);
use ZIB::Channel;
use ZIB::ChannelList;
use ZIB::Request;
use ZIB::RequestHandlers;


#
# User-serviceable parts below. ;)
#

my $server   = 'yourserver';
my $port     = 6667;
my $nick     = 'renameme';
my $user     = 'zib';
my $realname = 'zzznowman\'s IRC bot';


#
# Global state variables
#

my ($cont, $exitcode) = (1, 0);
my ($irc, $conn, $chans, @requests);
my %console_cmds;


#
# Actions
#

sub do_connect(;%)
{
	my (%args) = @_;

	if (%args)
	{
		notice("Connecting to %s...", $args{Server});
		$conn->connect(%args);
	}
	else
	{
		notice("Reconnecting...");
		$conn->connect();
	}

	my $socket = $conn->socket();

	# Detect dead connection.
	$socket->sockopt(SO_KEEPALIVE, 1);

	# XXX: SOL_TCP, TCP_KEEPIDLE not defined in standard Perl modules...
	# # Portability: Linux-specific: Reduce SO_KEEPALIVE timeout.
	# my $keepidle_old = $socket->getsockopt(SOL_TCP, TCP_KEEPIDLE);
	# $socket->setsockopt(SOL_TCP, TCP_KEEPIDLE, 5 * 60);
	# my $keepidle_new = $socket->getsockopt(SOL_TCP, TCP_KEEPIDLE);
	# info("Set TCP_KEEPIDLE from %s to %s", $keepidle_old, $keepidle_new);
}

sub do_quit($$)
{
	my ($ec, $quitmsg) = @_;

	# Allow zib-termination override by negative exit code.
	if ($ec >= 0)
	{
		# Arrange for zib termination on completion of disconnect.
		($cont, $exitcode) = (0, $ec);
	}

	notice("Quitting." . (defined($quitmsg) ? " [$quitmsg]" : ''));
	$conn->quit($quitmsg);
}

sub do_privmsg($$)
{
	my ($target, $line) = @_;

	debug("Sending PRIVMSG to %s: %s", $target, $line);
	$conn->privmsg($target, $line);
}

sub do_notice($$)
{
	my ($target, $line) = @_;

	debug("Sending NOTICE to %s: %s", $target, $line);
	$conn->notice($target, $line);
}


#
# State maintenance
#

sub register_console_cmds(&@)
{
	my ($coderef, @cmdnames) = @_;

	croak("can't register console command: $coderef is not a CODE ref, died")
	  unless (ref($coderef) eq 'CODE');

	foreach my $cmdname (@cmdnames)
	{
		$console_cmds{$cmdname} = $coderef;
	}
}

sub process_requests()
{
	my @newreqs = ();

	foreach my $request (@requests)
	{
		$request->do_request();
		
		unless ($request->ready())
		{
			# Push back request for later processing.
			push(@newreqs, $request);
			next;
		}

		# Send back reply.
		my $type  = $request->type();
		my $via   = $request->via();
		my $who   = $request->who();
		my @lines = $request->reply();
		my $path = sub { do_privmsg($via, "@_") };

		if ($via eq 'msg')
		{
			my $nick = raw2nick($who);

			$path = sub { do_notice($nick, "@_") };
		}
		elsif ($via eq 'notice')
		{
			my $nick = raw2nick($who);

			$path = sub { do_notice($nick, "@_") };
		}
		elsif ($via eq 'console')
		{
			$path = sub { print("Reply: @_\n") };
		}

		if (scalar(@lines))
		{
			info("Sending request reply for %s via %s to %s (%d lines)",
			     $type, $via, $who, scalar(@lines));
		}
		else
		{
			info("Sending empty request reply for %s via %s to %s",
			     $type, $via, $who);
			my @empty_replies = ZIB::Request->emptyreplies();
			push(@lines,
			     @empty_replies[int(rand(scalar(@empty_replies)))]);
		}

		foreach my $line (@lines)
		{
			$path->($line);
		}
	}

	@requests = @newreqs;
}

sub process_request_words($$@)
{
	my ($via, $who, $type, @args) = @_;

	# Ignore empty requests.
	unless (defined($type))
	{
		warning("Warning: process_request_words(): request type undefined, ignoring");
		return;
	}

	info("Request for %s via %s from %s [%s]", $type, $via, $who, "@args");

	eval
	{
		push(@requests, ZIB::Request->new($via, $who, $type, @args));
	};
	
	if (my $errmsg = $@)
	{
		chomp($errmsg);
		$errmsg =~ s#^(.*), died at ZIB/Request\.pm .*?$#$1#;
		info("Can't process request: %s", $errmsg);
		return undef;
	}
}

sub process_request_line($$$)
{
	my ($via, $who, $line) = @_;

	# Split into words; ignore lines consisting of whitespace only.
	my @words = split_line($line);
	return unless (scalar(@words));

	unless (defined(process_request_words($via, $who, @words)))
	{
		info("Input was: %s", $line);
	}
}


#
# Console command handlers
#

sub oncc_loglevel
{
	my ($cmd, @args) = @_;
	if    ($cmd eq 'loglevel') { setloglevel($args[0]) if (@args); }
	elsif ($cmd eq 'verbose')  { $loglevel++ }
	elsif ($cmd eq 'quiet')    { $loglevel-- }
	elsif ($cmd eq 'debug')    { setloglevel('debug') }
	elsif ($cmd eq 'nodebug')  { setloglevel('info')  }
	else { warning("Warning: oncc_loglevel(): invalid command name: $cmd"); }
	print("Log level is $loglevel\n");
}
register_console_cmds(\&oncc_loglevel, qw(loglevel verbose quiet debug nodebug));

sub oncc_reconnect
{
	my ($cmd, @args) = @_;

	# Performed action varies depending on whether still connected or not.
	if ($conn->connected())
	{
		# Just send a quit message, so that Net::IRC
		# will close the socket properly.
		do_quit(-1, @args ? "@args" : 'Reconnecting');
	}
	else
	{
		# (In case on_disconnect failed: ) Reconnect now.
		do_connect();
	}
}
register_console_cmds(\&oncc_reconnect, qw(reconnect reconn recon));

sub oncc_quit
{
	shift;
	my $quitmsg = join(' ', @_);
	do_quit(0, length($quitmsg) ? $quitmsg : undef);
}
register_console_cmds(\&oncc_quit, qw(quit));

sub oncc_join
{
	my $cmd = shift;
	my @channames = ();

	foreach my $arg (@_)
	{
		push(@channames, split(/,/, $arg));
	}

	if ($cmd eq 'join')
	{
		$chans->join_chans(ZIB::Channel->many_new(
		  {why => 'console', who => 'STDIN'}, @channames));
	}
	elsif ($cmd eq 'part')
	{
		$chans->part_chans(@channames);
	}
	else
	{
		warning("Warning: oncc_join(): invalid command name: $cmd");
	}
}
register_console_cmds(\&oncc_join, qw(join part));

sub oncc_request
{
	my ($cmd, @args) = @_;

	process_request_words('console', 'STDIN', @args);
}
register_console_cmds(\&oncc_request, qw(request req));

sub oncc_reload
{
	my ($cmd, @args) = @_;
	my $mod = 'ZIB::RequestHandlers';
	(my $modfile = $mod . '.pm') =~ s#::#/#g;

	warning("Trying to reload $mod...");

	unless (my $ret = do($INC{$modfile}))
	{
		if ($@)
		{
			chomp($@);
			foreach my $line (split(/\n/, $@))
			{
				error("Error: couldn't parse $modfile: %s", $line);
			}
		}
		else
		{
			if (!defined($ret) && length($!))
			{
				error("Error: couldn't read $modfile: %s", $!);
			}
			else
			{
				error("Error: couldn't reload $mod");
			}
		}

		return $ret;
	}

	warning("Reloaded $mod");
}
register_console_cmds(\&oncc_reload, qw(reload rehash));

sub oncc_channels
{
	print("Channels (name, want, have, why, who):\n");
	foreach my $chan ($chans->chans())
	{
		my ($name, $want, $have, $why, $who) =
		   ($chan->name(), $chan->want(), $chan->have(),
		    $chan->why(), $chan->who());
		printf("  %-14s  %d  %d  %-10s %s\n", $name,
		  $want, $have,
		  defined($why) ? $why : '?',
		  defined($who) ? $who : '?');
	}
}
register_console_cmds(\&oncc_channels, qw(channels chans));

sub oncc_status
{
	foreach my $cmd (qw(channels)) {
		$console_cmds{$cmd}->($cmd)
	}
}
register_console_cmds(\&oncc_status, qw(status stat));


#
# FileHandle handlers
#

sub onfh_stdin
{
	my ($fh) = @_;

	my $line = <$fh>;
	unless (defined($line))
	{
		$irc->removefh($fh);
		do_quit(1, "Lost terminal");
		return;
	}

	# Split into words; ignore lines consisting of whitespace only.
	my ($cmd, @args) = split_line($line);
	return unless (defined($cmd));

	# Look up command.
	my $cmdref = $console_cmds{$cmd};
	unless (defined($cmdref))
	{
		print("Invalid command: $cmd\n");
		return;
	}

	$cmdref->($cmd, @args);
}


#
# IRC handlers
#

sub on_error
{
	my ($self, $event) = @_;

	error("Got error from server:");
	$event->dump();
}

sub on_other
{
	my ($self, $event) = @_;

	notice("Got unidentified message from server:");
	$event->dump();
}

sub on_connect
{
	my ($self, $event) = @_;

	notice("Connected to server.");

	# Autojoin wanted channels.
	$chans->join_chans($chans->want());
}

sub on_disconnect
{
	my ($self, $event) = @_;

	# Reset channel "have" list.
	foreach my $chan ($chans->chans())
	{
		$chan->have(0);
	}

	my $args = join(' ', $event->args());
	$args =~ s/^ERROR\s+:Closing\s+Link:\s+\S+\s+(.*)$/$1/i;
	$args =~ s/^\((.*)\)$/$1/;
	notice("Disconnected from server. [%s]", $args);

	if ($cont)
	{
		# Connect again.
		do_connect();
	}
	else
	{
		# Don't continue any longer, terminate zib.
		notice("Exiting %d", $exitcode);
		exit $exitcode;
	}
}

sub on_leaving
{
	my ($self, $event) = @_;

	# This avoids errors during global destruction.
	return unless (defined($chans));

	# Reset channel "have" list here too
	# in case there was no disconnect event.
	foreach my $chan ($chans->chans())
	{
		$chan->have(0);
	}
}

sub on_join
{
	my ($self, $event) = @_;
	my $ourjoin = (raw2nick($event->from()) eq $self->nick());

	foreach my $channame ($event->to())
	{
		if ($ourjoin)
		{
			$chans->add_chan($channame)->have(1);
			info("Joined %s", $channame);
		}
		else
		{
			debug("Join: %s: %s", $channame, $event->from());
		}
	}
}

sub on_part
{
	my ($self, $event) = @_;
	my $ourpart = (raw2nick($event->from()) eq $self->nick());
	my $partmsg = join(' ', $event->args());

	foreach my $channame ($event->to())
	{
		if ($ourpart)
		{
			$chans->chan($channame)->have(0);
			info("Parted %s [%s]", $channame, $partmsg);
		}
		else
		{
			debug("Part: %s: %s [%s]",
			     $channame, $event->from(), $partmsg);
		}
	}
}

sub on_invite
{
	my ($self, $event) = @_;
	my @channames = $event->args();
	my $who = $event->from();

	info("Invited to %s by %s", join(',', @channames), $who);

	$chans->join_chans(ZIB::Channel->many_new(
	  {why => 'invite', who => $who}, @channames));
}

sub on_nick
{
	my ($self, $event) = @_;

	debug("Nick: %s from %s", $event->to(), $event->from());
}

sub on_quit
{
	my ($self, $event) = @_;

	debug("Quit: %s [%s]", $event->from(), join(' ', $event->args()));
}

sub on_public
{
	my ($self, $event) = @_;
	my @channames = $event->to();
	my $line = join(' ', $event->args());
	my $who = $event->from();

	debug("Public: %s: %s: %s", join(',', @channames), $who, $line);

	# Ignore own messages.
	return if (raw2nick($who) eq $self->nick());

	# Ignore messages without command prefix.
	return unless ($line =~ /^\.(\S.*)$/);

	foreach my $channame (@channames)
	{
		process_request_line($channame, $who, $1);
	}
}

sub on_msg
{
	my ($self, $event) = @_;
	my $line = join(' ', $event->args());
	my $who = $event->from();

	debug("Msg: %s: %s", $who, $line);
	process_request_line('msg', $who, $line);
}

sub on_notice
{
	my ($self, $event) = @_;
	my @targetnames = $event->to();
	my $line = join(' ', $event->args());
	my $who = $event->from();
	my $ournotice = (scalar(@targetnames) == 1 &&
	                 raw2nick($targetnames[0]) eq $self->nick());

	# Server notice?
	if ($who =~ /^[^!@]*\./)
	{
		if ($ournotice)
		{
			info("Server notice: %s: %s", $who, $line);
		}
		else
		{
			info("Server notice for %s: %s: %s",
			     join(',', @targetnames), $who, $line);
		}
	}
	else
	{
		if ($ournotice)
		{
			info("Notice: %s: %s", $who, $line);
		}
		else
		{
			debug("Notice for %s: %s: %s",
			      join(',', @targetnames), $who, $line);
		}
	}
}


#
# Prepare Net::IRC.
#

# Create global IRC object.
$irc = Net::IRC->new();

# Accept input from stdin.
$irc->addfh(\*STDIN, \&onfh_stdin, 'r');

# Create global connection object.
$conn = $irc->newconn();

# Uncomment for full traffic logging.
#$conn->debug(1);


# Set handlers.
# TODO: mode, topic, kick, kill, umode, CTCP
# TODO: bannedfromchan, nosuchnick

$conn->add_handler('error', \&on_error, 1);
$conn->add_handler('other', \&on_other, 1);

$conn->add_handler('welcome',    \&on_connect,    2);
$conn->add_handler('disconnect', \&on_disconnect, 0);
$conn->add_handler('leaving',    \&on_leaving,    1);

$conn->add_handler('join',   \&on_join,   2);
$conn->add_handler('part',   \&on_part,   2);
$conn->add_handler('invite', \&on_invite, 2);

$conn->add_handler('nick', \&on_nick, 2);
$conn->add_handler('quit', \&on_quit, 2);

$conn->add_handler('public', \&on_public, 2);
$conn->add_handler('msg',    \&on_msg,    2);
$conn->add_handler('notice', \&on_notice, 2);


# Initialize global channel list.
$chans = ZIB::ChannelList->new($conn);


# Connect to server.
do_connect(Server   => $server,
           Port     => $port,
           Nick     => $nick,
           Username => $user,
           Ircname  => $realname);


#
# Enter event loop.
#

while (1)
{
	log_checkstate();
	$irc->do_one_loop();
	process_requests();
}
